@@ -10,33 +10,57 @@ from itertools import compress
 # User
 from ..estimator import (boosting)
 from .sanity_check import check_estimator
+from .exceptions import WrongInputException
 
 #------------------------------------------------------------------------------
 # Model tools
 #------------------------------------------------------------------------------
-def compose_model(name, perform_estimator_check=True, verbose=True):
+def compose_model(estimator_name, perform_estimator_check=True, verbose=True):
         
         # Modules from scikit-learn used for regression
         SCIKIT_REGRESSION_MODULES = ["dummy","ensemble","gaussian_process",
                                      "isotonic","kernel_ridge","linear_model",
                                      "neighbors","neural_network","svm","tree",]
-
-        # Check if 'name' is an attribute of any Scikit-learn module
-        scikit_regression_modules_check = [hasattr(eval(lib), name) for lib in SCIKIT_REGRESSION_MODULES]
-        is_scikit = any(scikit_regression_modules_check)
         
-        # Compose model
-        if is_scikit:
-            # Locate library
-            scikit_module = list(compress(SCIKIT_REGRESSION_MODULES, scikit_regression_modules_check))[0]
-            if verbose:
-                print(f"Algorithm '{name}' is part of Scikit-learn, namely an attribute of: '{scikit_module}'")
-                            
-            # Instantiate model
-            estimator = eval(scikit_module+"."+name+"()")
+        # Internal modules used for regression
+        NONSCIKIT_REGRESSION_MODULES = ["boosting",]
+        
+        REGRESSION_MODULES = [SCIKIT_REGRESSION_MODULES,NONSCIKIT_REGRESSION_MODULES]
+        
+        # Count the number of modules in which we find the model name
+        n_modules_found_in = 0
+        
+        for reg_module in REGRESSION_MODULES:
             
-        else:
-            raise NotImplementedError()
+            # Check if 'estimator_name' is an attribute of any module
+            module_check = [hasattr(eval(lib), estimator_name) for lib in reg_module]
+        
+            is_found = any(module_check)
+    
+            # Compose model            
+            if is_found:
+                
+                # Check counter
+                if n_modules_found_in>1:
+                    raise Exception(f"""Algorithm '{estimator_name}' is part of multiple modules, namely at least two of
+                                    \n{SCIKIT_REGRESSION_MODULES}
+                                    \n{NONSCIKIT_REGRESSION_MODULES}
+                                    """)
+                
+                # Locate module
+                module = list(compress(reg_module, module_check))[0]
+
+                # Instantiate model
+                estimator = eval(module+"."+estimator_name+"()")
+
+                # Increase model counter
+                n_modules_found_in += 1
+
+        if n_modules_found_in==0:
+            raise Exception(f"""Algorithm '{estimator_name}' is NOT part any module, namely neither of
+                            \n{SCIKIT_REGRESSION_MODULES}
+                            \n{NONSCIKIT_REGRESSION_MODULES}
+                            """)
 
         # Model checks
         if perform_estimator_check:
@@ -61,8 +85,23 @@ def update_params(old_param, new_param, errors="raise"):
                 raise Exception(f"Parameters {k} not recognized as a default parameter for this estimator")
             else:
                 pass
+            
     return updated_param
 
+def isinstance_double(obj, class_or_tuple, how="all"):
+    ALLOWED_HOW = ["all", "any"]
+    
+    if how=="all":
+        is_instance = isinstance(obj,class_or_tuple) and (obj.__class__.__name__==class_or_tuple.__name__)
+    elif how=="any":
+        is_instance = isinstance(obj,class_or_tuple) or (obj.__class__.__name__==class_or_tuple.__name__)    
+    else:
+        raise WrongInputException(input_name="how",
+                                  provided_input=how,
+                                  allowed_inputs=ALLOWED_HOW)
+    
+    return is_instance
+
 
 def get_param_grid_from_estimator(estimator):
     """
@@ -73,7 +112,7 @@ def get_param_grid_from_estimator(estimator):
     #--------------------------------------------------------------------------
     # from ..estimator.boosting
     #--------------------------------------------------------------------------        
-    if isinstance(estimator, boosting.XGBRegressor):
+    if isinstance_double(estimator, boosting.XGBRegressor):
         param_grid = {
             "n_estimators":[100,200,500],
             "max_depth":[None,2,4,8,16],
@@ -99,7 +138,7 @@ def get_param_grid_from_estimator(estimator):
             "importance_type":'gain',
             }
     
-    elif isinstance(estimator, boosting.LGBMegressor):
+    elif isinstance_double(estimator, boosting.LGBMegressor):
         param_grid = {
             "boosting_type":'gbdt',
             "num_leaves":[31,11,21,41],
@@ -123,7 +162,7 @@ def get_param_grid_from_estimator(estimator):
     #--------------------------------------------------------------------------
     # from sklearn.dummy
     #--------------------------------------------------------------------------
-    elif isinstance(estimator, dummy.DummyRegressor):
+    elif isinstance_double(estimator, dummy.DummyRegressor):
         param_grid = {
             "strategy":["mean","median","quantile","constant"],
             "constant":None,
@@ -133,15 +172,15 @@ def get_param_grid_from_estimator(estimator):
     #--------------------------------------------------------------------------
     # from sklearn.ensemble
     #--------------------------------------------------------------------------
-    # elif isinstance(estimator, ensemble.AdaBoostRegressor):
+    # elif isinstance_double(estimator, ensemble.AdaBoostRegressor):
     #     raise NotImplementedError()
         
-    # elif isinstance(estimator, ensemble.BaggingRegressor):
+    # elif isinstance_double(estimator, ensemble.BaggingRegressor):
     #     raise NotImplementedError()
         
-    elif isinstance(estimator, ensemble.ExtraTreesRegressor):
+    elif isinstance_double(estimator, ensemble.ExtraTreesRegressor):
         param_grid = {
-            "n_estimators":500,                                                # Default 100 
+            "n_estimators":200,                                                # Default 100
             'criterion':'squared_error',
             'max_depth':[None,2,4,8,16],
             'min_samples_split':[2,4,8,16],
@@ -156,7 +195,7 @@ def get_param_grid_from_estimator(estimator):
             'max_samples':None
             }   
 
-    elif isinstance(estimator, ensemble.GradientBoostingRegressor):
+    elif isinstance_double(estimator, ensemble.GradientBoostingRegressor):
         param_grid = {
             "loss":'squared_error',
             'learning_rate':[0.1,0.8, 0.5, 0.3, 0.01],
@@ -178,15 +217,15 @@ def get_param_grid_from_estimator(estimator):
             'ccp_alpha':0.0
             }
         
-    elif isinstance(estimator, ensemble.RandomForestRegressor):
+    elif isinstance_double(estimator, ensemble.RandomForestRegressor):
         param_grid = {
-            "n_estimators":500,                                                # Default 100 
+            "n_estimators":200,                                                # Default 100
             "criterion":'squared_error',
             "max_depth":[None,2,4,8,16],
             "min_samples_split":[2,4,8,16],
             "min_samples_leaf":[1,2,4,8],
             "min_weight_fraction_leaf":0.0,
-            "max_features":['sqrt',1/4,1/3,1/2,2/3,'log2','auto'],             # Default 'auto'
+            "max_features":['sqrt',1/4,1/3,1/2,2/3,'log2','auto'],             # Default 'auto'    
             "max_leaf_nodes":None,
             "min_impurity_decrease":0.0,
             "bootstrap":True,
@@ -195,44 +234,43 @@ def get_param_grid_from_estimator(estimator):
             "random_state":None,
             "verbose":0,
             "warm_start":False,
-            "ccp_alpha":0.0,
+            # "ccp_alpha":0.0, # Leads to RuntimeError: Cannot clone object RandomForestRegressor(...), as the constructor either does not set or modifies parameter ccp_alpha
             "max_samples":None
             }
 
-    # elif isinstance(estimator, ensemble.HistGradientBoostingRegressor):
+    # elif isinstance_double(estimator, ensemble.HistGradientBoostingRegressor):
     #     raise NotImplementedError()    
 
     #--------------------------------------------------------------------------
     # from sklearn.gaussian_process
     #--------------------------------------------------------------------------
-    # elif isinstance(estimator, gaussian_process.GaussianProcessRegressor):
+    # elif isinstance_double(estimator, gaussian_process.GaussianProcessRegressor):
     #     raise NotImplementedError()    
 
     #--------------------------------------------------------------------------
     # from sklearn.isotonic
     #--------------------------------------------------------------------------
-    # elif isinstance(estimator, isotonic.IsotonicRegression):
+    # elif isinstance_double(estimator, isotonic.IsotonicRegression):
     #     raise NotImplementedError()    
     
     #--------------------------------------------------------------------------
     # from sklearn.kernel_ridge
     #--------------------------------------------------------------------------
-    # elif isinstance(estimator, kernel_ridge.KernelRidge):
+    # elif isinstance_double(estimator, kernel_ridge.KernelRidge):
     #     raise NotImplementedError()    
     
     #--------------------------------------------------------------------------
     # from sklearn.linear_model
     #--------------------------------------------------------------------------
-    elif isinstance(estimator, linear_model.LinearRegression):
+    elif isinstance_double(estimator, linear_model.LinearRegression):
         param_grid = {
             "fit_intercept":True,
-            "normalize":'deprecated',
             "copy_X":True,
             "n_jobs":None,
             "positive":False
             }
                 
-    elif isinstance(estimator, linear_model.Ridge):
+    elif isinstance_double(estimator, linear_model.Ridge):
         param_grid = {
             'alpha':[1.0,0.001, 0.1, 10, 100, 1000, 10000], 
             'fit_intercept':True,
@@ -244,7 +282,8 @@ def get_param_grid_from_estimator(estimator):
             'random_state':None
             }
         
-    elif isinstance(estimator, linear_model.RidgeCV):
+
+    elif isinstance_double(estimator, linear_model.RidgeCV):
         param_grid = {
             'alphas':(1.0,0.001, 0.1, 10, 100, 1000, 10000),
             'fit_intercept':True,
@@ -254,23 +293,23 @@ def get_param_grid_from_estimator(estimator):
             'alpha_per_target':False
             }
         
-    # elif isinstance(estimator, linear_model.SGDRegressor):
+    # elif isinstance_double(estimator, linear_model.SGDRegressor):
     #     raise NotImplementedError()    
         
-    elif isinstance(estimator, linear_model.ElasticNet):
+    elif isinstance_double(estimator, linear_model.ElasticNet):
         param_grid = {
             'alpha':np.exp(np.linspace(start=np.log(100), stop=np.log(100*0.000001), num=1000)),
             'l1_ratio':[0.5,1/4,3/4,1],
             'fit_intercept':True,
             'precompute':False,
-            'max_iter':1000,
+            'max_iter':10,
             'tol':0.0001,
             'warm_start':False,
             'positive':False,
             'selection':'cyclic'
             }
 
-    elif isinstance(estimator, linear_model.ElasticNetCV):
+    elif isinstance_double(estimator, linear_model.ElasticNetCV):
         param_grid = {
             'l1_ratio':[0.5,1/4,3/4,1],
             'eps':0.001,
@@ -284,7 +323,7 @@ def get_param_grid_from_estimator(estimator):
             'selection':'cyclic'
             }
 
-    elif isinstance(estimator, linear_model.Lars):
+    elif isinstance_double(estimator, linear_model.Lars):        
         param_grid = {
             'fit_intercept':True,
             'precompute':'auto',
@@ -295,7 +334,7 @@ def get_param_grid_from_estimator(estimator):
             'jitter':None,
             }
 
-    elif isinstance(estimator, linear_model.LarsCV):
+    elif isinstance_double(estimator, linear_model.LarsCV):
         param_grid = {
             'fit_intercept':True,
             'max_iter':500,
@@ -305,7 +344,7 @@ def get_param_grid_from_estimator(estimator):
             'copy_X':True
             }
         
-    elif isinstance(estimator, linear_model.Lasso):
+    elif isinstance_double(estimator, linear_model.Lasso):
         param_grid = {
             'alpha':np.exp(np.linspace(start=np.log(100), stop=np.log(100*0.000001), num=1000)),
             'fit_intercept':True,
@@ -318,7 +357,7 @@ def get_param_grid_from_estimator(estimator):
             'selection':'cyclic'
             }
 
-    elif isinstance(estimator, linear_model.LassoCV):
+    elif isinstance_double(estimator, linear_model.LassoCV):
         param_grid = {
             'eps':0.001,
             'n_alphas':1000,
@@ -331,7 +370,7 @@ def get_param_grid_from_estimator(estimator):
             'selection':'cyclic'
             }
 
-    elif isinstance(estimator, linear_model.LassoLars):
+    elif isinstance_double(estimator, linear_model.LassoLars):
         param_grid = {
             'alpha':np.exp(np.linspace(start=np.log(100), stop=np.log(100*0.000001), num=1000)),
             'fit_intercept':True,
@@ -344,7 +383,7 @@ def get_param_grid_from_estimator(estimator):
             'jitter':None,
             }
 
-    elif isinstance(estimator, linear_model.LassoLarsCV):
+    elif isinstance_double(estimator, linear_model.LassoLarsCV):
         param_grid = {
             'fit_intercept':True,
             'max_iter':500,
@@ -355,25 +394,25 @@ def get_param_grid_from_estimator(estimator):
             'positive':False
             }
         
-    # elif isinstance(estimator, linear_model.OrthogonalMatchingPursuit):
+    # elif isinstance_double(estimator, linear_model.OrthogonalMatchingPursuit):
     #     raise NotImplementedError()            
         
-    # elif isinstance(estimator, linear_model.ARDRegression):
+    # elif isinstance_double(estimator, linear_model.ARDRegression):
     #     raise NotImplementedError()           
 
-    # elif isinstance(estimator, linear_model.BayesianRidge):
+    # elif isinstance_double(estimator, linear_model.BayesianRidge):
     #     raise NotImplementedError()           
 
-    # elif isinstance(estimator, linear_model.HuberRegressor):
+    # elif isinstance_double(estimator, linear_model.HuberRegressor):
     #     raise NotImplementedError()           
 
-    # elif isinstance(estimator, linear_model.QuantileRegressor):
+    # elif isinstance_double(estimator, linear_model.QuantileRegressor):
     #     raise NotImplementedError()           
 
-    # elif isinstance(estimator, linear_model.RANSACRegressor):
+    # elif isinstance_double(estimator, linear_model.RANSACRegressor):
     #     raise NotImplementedError()           
 
-    # elif isinstance(estimator, linear_model.TheilSenRegressor):
+    # elif isinstance_double(estimator, linear_model.TheilSenRegressor):
     #     raise NotImplementedError()                   
         
     #--------------------------------------------------------------------------
@@ -383,7 +422,7 @@ def get_param_grid_from_estimator(estimator):
     #--------------------------------------------------------------------------
     # from sklearn.neural_network
     #--------------------------------------------------------------------------
-    elif isinstance(estimator, neural_network.MLPRegressor):
+    elif isinstance_double(estimator, neural_network.MLPRegressor):
         param_grid = {
             'hidden_layer_sizes':[(100,), (2,), (2,4), (2,4,8,), (2,4,8,16,), (2,4,8,16,32),
                                           (4,), (4,8), (4,8,16), (4,8,16,32),
